from flask import current_app, request, abort, jsonify
from datetime import datetime
from . import forecast as forecast_
from . import episode as episode_
from .autocomplete import autocomplete as autocomplete_
import pytz
from dateutil import parser

@current_app.route('/forecast')
def forecast():
    insee = request.args.get('insee')
    zone = pytz.timezone('Europe/Paris')
    date_ = request.args.get('date')

    try:
        result = forecast_(insee, date_)
    except KeyError as e:
        current_app.logger.error(f'INSEE {insee} not found')
        current_app.logger.error(e)
        abort(404)

    return jsonify(result)

@current_app.route('/episodes')
def episodes():
    insee = request.args.get('insee')
    zone = pytz.timezone('Europe/Paris')
    date_ = request.args.get('date') or str(datetime.now(tz=zone).date().isoformat())

    try:
        result = episode_(insee, date_)
    except KeyError as e:
        current_app.logger.error(f'INSEE {insee} not found')
        current_app.logger.error(e)
        abort(404)

    return jsonify(result)

@current_app.route('/autocomplete')
def autocomplete():
    q = request.args.get('q')

    return jsonify(autocomplete_(q))



from flask import Flask
from flask_manage_webpack import FlaskManageWebpack
from flask_cors import CORS
from flask_migrate import Migrate
from datetime import datetime
import pytz
import os

def create_app(test_config=None):
    app = Flask(
        __name__,
        instance_relative_config=True,
        static_url_path=''
    )
    app.config.from_mapping(
        SECRET_KEY=os.getenv('SECRET_KEY', 'dev'),
        SQLALCHEMY_DATABASE_URI=os.getenv('SQLALCHEMY_DATABASE_URI') or os.getenv('POSTGRESQL_ADDON_URI'),
        SQLALCHEMY_TRACK_MODIFICATIONS=False
    )
    CORS(app, send_wildcard=True)

    manage_webpack = FlaskManageWebpack()
    manage_webpack.init_app(app)

    from .models import db
    db.init_app(app)
    migrate = Migrate(app, db)

    with app.app_context():
        import indice_pollution.api
        import indice_pollution.web
        import indice_pollution.history

    return app

def make_resp(r, result):
    return {
        "data": result,
        "metadata": {
            "region": {
                "nom": r.__module__.split(".")[-1],
                "website": r.website
            }
        }
    }

def forecast(insee, date_=None, force_from_db=False):
    from .regions.solvers import region
    date_ = date_ or today()
    forecast = region(insee).Forecast()
    result = forecast.get(date_=date_, insee=insee, force_from_db=force_from_db)
    return make_resp(forecast, result)

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def bulk(insee_region_names, date_=None, fetch_episodes=False):
    from indice_pollution.history.models import IndiceHistory, EpisodeHistory
    from .regions.solvers import region as get_region
    date_ = date_ or today()

    insees = set(insee_region_names.keys())
    close_insees = set()
    for insee in insees:
        region = get_region(region_name=insee_region_names[insee])
        close_insee = region.Forecast().get_close_insee(insee)
        close_insees.add(close_insee)
    insees.update(close_insees)

    indices = dict()
    episodes = dict()
    for chunk in chunks(list(insees), 10):
        indices.update(
            {i.insee: [i.features] for i in IndiceHistory.get_bulk(date_, chunk)}
        )
        if fetch_episodes:
            episodes.update(
                {i.insee: [i.features] for i in EpisodeHistory.get_bulk(date_, chunk)}
            )
    for insee in insee_region_names.keys():
        if insee in indices:
            continue
        region = get_region(region_name=insee_region_names[insee])
        f = region.Forecast()
        close_insee = f.get_close_insee(insee)
        if close_insee in indices:
            indices[insee] = indices[close_insee]
            continue
        indices[insee] = f.get(date_=date_, insee=insee, force_from_db=False)
    if fetch_episodes:
        for insee in insee_region_names.keys():
            if insee in episodes:
                continue
            region = get_region(region_name=insee_region_names[insee])
            e = region.Episode()
            close_insee = e.get_close_insee(insee)
            if close_insee in episodes:
                episodes[insee] = episodes[close_insee]
                continue
            episodes[insee] = e.get(date_=date_, insee=insee, force_from_db=False)
    to_return = {
        insee: {
            "forecast": make_resp(
                get_region(region_name=insee_region_names[insee]).Forecast(),
                indices[insee]
               ),
        }
        for insee in insee_region_names.keys()
    }
    if fetch_episodes:
        for insee in insee_region_names:
            to_return[insee].update({
                "episode": make_resp(
                    get_region(region_name=insee_region_names[insee]).Episode(),
                    episodes[insee]
                )
                }
            )
    return to_return

def today():
    zone = pytz.timezone('Europe/Paris')
    return datetime.now(tz=zone).date()

def episode(insee, date_=None):
    from .regions.solvers import region
    zone = pytz.timezone('Europe/Paris')
    date_ = date_ or today()
    episode = region(insee).Episode()
    return {
        "data": episode.get(date_=date_, insee=insee),
        "metadata": {
            "region": {
                "nom": episode.__module__.split(".")[-1],
                "website": episode.website
            }
        }
    }



from indice_pollution.regions.solvers import insee_list
from indice_pollution import forecast as forecast_, today
from indice_pollution.models import db
from indice_pollution.history.models import IndiceHistory
from flask import current_app
from datetime import timedelta

@current_app.cli.command('generate-history')
def generate_history():
    date_ = today()
    for insee in insee_list():
        current_app.logger.info(f'Getting history for {insee}')
        forecast_(insee, date_, force_from_db=True)

@current_app.cli.command('prune-history')
def prune_history():
    date_ = today() - timedelta(days=7)
    IndiceHistory.query.filter(IndiceHistory.date_ <= date_).delete()
    current_app.logger.info(f'Data before {date_} is removed from history')



from . import ServiceMixin, ForecastMixin, EpisodeMixin
from datetime import timedelta
from dateutil.parser import parse

class Service(ServiceMixin):
    date_format = '%Y-%m-%d'
    website = 'https://www.atmo-hdf.fr/'
    insee_list = [
        '59350', '59183', '59178', '62041', '62160', '59606', '02691', '60175',
        '62765', '62193', '59392', '80021', '62119'
    ]

class Forecast(Service, ForecastMixin):
    url = 'https://services8.arcgis.com/rxZzohbySMKHTNcy/arcgis/rest/services/ind_hdf_agglo/FeatureServer/0/query'

class Episode(Service, EpisodeMixin):
    url = 'https://services8.arcgis.com/rxZzohbySMKHTNcy/arcgis/rest/services/alrt3j_hdf/FeatureServer/0/query?outFields='


from indice_pollution.regions.solvers import insee_list
import requests
import logging
import time
import pytz
from datetime import datetime
from dateutil.parser import parse
from sqlalchemy import exc
from indice_pollution.history.models import IndiceHistory, EpisodeHistory
from indice_pollution.models import db

class ServiceMixin(object):
    url = ""
    epci_list = []
    insee_epci = dict()
    outfields = ['date_ech', 'valeur', 'qualif', 'val_no2', 'val_so2',
     'val_o3', 'val_pm10', 'val_pm25'
    ]

    attributes_key = 'attributes'
    use_dateutil_parser = False

    HTTPAdapter = requests.adapters.HTTPAdapter

    def get_one_attempt(self, url, params):
        s = requests.Session()
        s.mount('https://', self.HTTPAdapter())
        try:
            r = s.get(url, params=params)
        except requests.exceptions.ConnectionError as e:
            logging.error(f'Impossible de se connecter à {url}')
            logging.error(e)
            return None
        except requests.exceptions.SSLError as e:
            logging.error(f'Erreur ssl {url}')
            logging.error(e)
            return None
        try:
            r.raise_for_status()
        except requests.HTTPError as e:
            logging.error(f'Erreur HTTP: {e}')
            return None
        return r

    def get_multiple_attempts(self, url, params, attempts=0):
        r = self.get_one_attempt(url, params)
        if r:
            features = self.features(r)
        else:
            features = []
        if attempts >= 3 or len(features) > 0:
            return features
        else:
            time.sleep(0.5 * (attempts + 1))
            return self.get_multiple_attempts(url, params, attempts+1)

    def get(self, date_, insee, attempts=0, force_from_db=False):
        to_return = []
        insee = self.get_close_insee(insee)
        if force_from_db:
            indice = self.HistoryModel.get(date_, insee)
            if indice:
                return [indice.features]
        if not to_return:
            to_return = self.get_no_cache(date_, insee, attempts)
        if not to_return:
            to_return = self.HistoryModel.get_after(date_, insee)
        if not any(map(lambda r: (r['date'] if 'date' in r else r['date_dif']) == str(date_), to_return)):
            if hasattr(self, "get_from_scraping"):
                to_return = self.get_from_scraping(to_return, date_, insee)
        if to_return:
            for v in to_return:
                if not v.get('indice'):
                    continue
                indice = self.HistoryModel.get_or_create(v['date'], insee)
                indice.features = v
                db.session.commit()
        return to_return

    def get_no_cache(self, date_, insee, attempts=0):
        features = self.get_multiple_attempts(self.url, self.params(date_, insee), attempts)
        return list(filter(lambda s: s is not None, map(self.getter, features)))

    def features(self, r):
        return r.json()['features']

    def where(self, date_, insee):
        zone = insee if not self.insee_epci else self.insee_epci[insee]
        return f"(date_ech >= '{date_}') AND (code_zone='{zone}')"

    def params(self, date_, insee):
        return {
            'where': self.where(date_, insee),
            'outFields': ",".join(self.outfields),
            'f': 'json',
            'outSR': '4326'
        }

    @property
    def insee_list(self):
        return [] if not self.insee_epci else self.insee_epci.keys()

    def get_close_insee(self, insee):
        if insee in self.insee_list:
            return insee
        departement = insee[:2]
        try:
            return next(pref_insee for pref_insee in self.insee_list if pref_insee[:2] == departement)
        except StopIteration:
            logging.error(f'Impossible de trouver le code insee de la préfecture de {insee}')
            raise KeyError

    def getter(self, feature):
        attributes = feature[self.attributes_key]

        dt = self.date_getter(attributes)
        return {
            **{
                'indice': attributes['valeur'],
                'date': str(dt.date())
            },
            **{k: attributes[k] for k in self.outfields if k in attributes}
        }

    def date_getter(self, attributes):
        if not self.use_dateutil_parser:
            zone = pytz.timezone('Europe/Paris')
            return datetime.fromtimestamp(attributes['date_ech']/1000, tz=zone)
        else:
            return parse(attributes['date_ech'])

    def date_parser(self, date_):
        if date_ is None:
            return
        if not self.use_dateutil_parser:
            zone = pytz.timezone('Europe/Paris')
            return datetime.fromtimestamp(date_/1000, tz=zone)
        else:
            return parse(date_)


class ForecastMixin(ServiceMixin):
    HistoryModel = IndiceHistory
    outfields = ['date_ech', 'valeur', 'qualif', 'val_no2', 'val_so2',
     'val_o3', 'val_pm10', 'val_pm25'
    ]

    def getter(self, feature):
        attributes = feature[self.attributes_key]

        dt = self.date_parser(attributes['date_ech'])
        return {
            **{
                'indice': attributes['valeur'],
                'date': str(dt.date())
            },
            **{k: attributes[k] for k in self.outfields if k in attributes}
        }


class EpisodeMixin(ServiceMixin):
    HistoryModel = EpisodeHistory
    outfields = ['date_ech', 'lib_zone', 'code_zone', 'date_dif', 'code_pol',
     'lib_pol', 'etat', 'com_court', 'com_long']
    
    def getter(self, feature):
        attributes = feature[self.attributes_key]

        if attributes['etat'] == 'PAS DE DEPASSEMENT':
            return None

        date_ech = self.date_parser(attributes['date_ech'])
        date_dif = self.date_parser(attributes.get('date_dif'))
        return {
            **{k: attributes[k] for k in self.outfields if k in attributes},
            **{
                'date_dif': str(date_dif.date()) if date_dif else None,
                'date_ech': str(date_ech.date())
            }
        }

    def centre(self, insee):
        r = requests.get(
            f'https://geo.api.gouv.fr/communes/{insee}',
            params={"fields": "centre", "format": "json", "geometry": "centre"}
        )
        r.raise_for_status()
        return r.json()['centre']['coordinates']

    def params(self, date_, insee):
        centre = self.centre(insee)

        return {
            'where': '',
            'outfields': self.outfields,
            'f': 'json',
            'geometry': f'{centre[0]},{centre[1]}',
            'inSR': '4326',
            'outSR': '4326',
            'geometryType': 'esriGeometryPoint'
        }



from . import ForecastMixin, EpisodeMixin
import requests
from bs4 import BeautifulSoup
from datetime import datetime

class Service(object):
    website = 'https://www.atmo-auvergnerhonealpes.fr/'
    insee_list = [
        '73065', '74081', '43157', '69123', '73248', '03190', '74012', '38544', '73011',
        '63300', '01053', '74056', '03310', '03185', '42187', '15014', '38185', '74010',
        '42218', '26362', '38053', '01328', '63113'
    ]

    def get_close_insee(self, insee):
        return insee

class Forecast(Service, ForecastMixin):
    url = 'https://services3.arcgis.com/o7Q3o5SkiSeZD5LK/arcgis/rest/services/ind_atmo_aura/FeatureServer/0/query'

    COLOR_TO_QUALIF = {
        "0": "tres_bon",
        "21": "bon",
        "31": "bon",
        "41": "bon",
        "51": "moyen",
        "61": "mediocre",
        "71": "mediocre",
        "81": "mediocre",
        "91": "mauvais",
        "100": "tres_mauvais"
    }

    def get_from_scraping(self, previous_results, date_, insee):
        r = requests.get(f'https://www.atmo-auvergnerhonealpes.fr/monair/commune/{insee}')
        r.raise_for_status()
        soup = BeautifulSoup(r.text, 'html.parser')

        controls = soup.find_all('div', class_='day-controls')
        days = controls[0].find_all('a', class_='raster-control-link')

        return [
            {
                "date": str(datetime.fromtimestamp(int(day.attrs.get('data-rasterid'))).date()),
                "indice": int(int(day.attrs.get('data-index'))/10),
                "qualif": self.COLOR_TO_QUALIF[day.attrs.get('data-color')]
            }
            for day in days
        ]


class Episode(Service, EpisodeMixin):
    url = 'https://services3.arcgis.com/o7Q3o5SkiSeZD5LK/arcgis/rest/services/Episodes%20de%20pollution%20pr%C3%A9vus%20ou%20constat%C3%A9s/FeatureServer/0/query'



from . import ServiceMixin, ForecastMixin, EpisodeMixin

class Service(ServiceMixin):
    website =  'https://www.qualitaircorse.org/'
    insee_list = ['2B096', '2B033', '2A004']

class Forecast(Service, ForecastMixin):
    url = 'https://services9.arcgis.com/VQopoXNvUqHYZHjY/arcgis/rest/services/ind_atmo_corse/FeatureServer/0/query'

class Episode(Service, EpisodeMixin):
    url = 'https://services9.arcgis.com/VQopoXNvUqHYZHjY/arcgis/rest/services/Alrt3j_corse/FeatureServer/0/query'


from . import ForecastMixin, EpisodeMixin
from datetime import timedelta, datetime

class Service(object):
    website = 'https://www.atmosud.org/'
    attributes_key = 'properties'
    use_dateutil_parser = True
    fr_date_format = '%Y-%m-%dT00:00:00Z'
    insee_list = ['06029', '06088', '13001', '13055', '83137', '84007']

class Episode(Service, EpisodeMixin):
    url = 'https://geoservices.atmosud.org/geoserver/alrt_sudpaca_dep/ows'

    def params(self, date_, insee):
        centre = self.centre(insee)

        return {
            'service': 'WFS',
            'version': '1.1.0',
            'request': 'GetFeature',
            'typeName': 'alrt_sudpaca_dep:alrt3j_sudpaca',
            'outputFormat': 'json',
            'geometry': f'{centre[0]},{centre[1]}',
            'inSR': '4326',
            'geometryType': 'esriGeometryPoint',
        }

class Forecast(Service, ForecastMixin):
    url = 'https://geoservices.atmosud.org/geoserver/ind_sudpaca/ows'

    @classmethod
    def params(cls, date_, insee):
        tomorrow_date = date_ + timedelta(days=1)

        fr_date = date_.strftime(cls.fr_date_format)
        fr_tomorrow = tomorrow_date.strftime(cls.fr_date_format)

        return {
            'service': 'WFS',
            'version': '1.1.0',
            'request': 'GetFeature',
            'typeName': 'ind_sudpaca:ind_sudpaca_agglo',
            'CQL_FILTER': f"code_zone='{insee}' AND (date_ech='{fr_date}' OR date_ech='{fr_tomorrow}')",
            'outputFormat': 'json'
        }
    def date_getter(self, attributes):
        return datetime.strptime(attributes['date_ech'], self.fr_date_format)
